﻿using System;
using System.Drawing;

namespace lcdS1D15G10com
{
	class ImageCtrl
	{
		public int ImgW { get; private set; }
		public int ImgH { get; private set; }
		public bool IsError { get; private set; }
		private Bitmap img;

		//コンストラクタ
		public ImageCtrl(string filename)
		{
			try
			{
				img = new Bitmap(filename);
				ImgW = img.Width;
				ImgH = img.Height;
				IsError = false;
			}
			catch { img = null; ImgW = ImgH = 0; IsError = true; }
		}

		//減色した画像を保存する
		private void SaveImage(TRgbColor[] rgbColor)
		{
			Bitmap saveImg = new Bitmap(ImgW, ImgH);
			Color pixColor;
			int rr, gg, bb;
			for (int i = 0, y = 0; y < ImgH; y++)
			{
				for (int x = 0; x < ImgW; x++, i++)
				{
					rr = rgbColor[i].rr * 256 / 16;
					gg = rgbColor[i].gg * 256 / 16;
					bb = rgbColor[i].bb * 256 / 16;
					pixColor = Color.FromArgb(rr, gg, bb);
					saveImg.SetPixel(x, y, pixColor);
				}
			}

			if (!string.IsNullOrEmpty(App.savePath))
			{
				DateTime dt = DateTime.Now;
				string filename = App.savePath + dt.ToFileTime() + ".png";
				saveImg.Save(filename);
			}
		}

		//縦横サイズをリサイズする
		//引数:	(width=n, height=0) または (width=0, height=n) または (width=m, height=n)
		public void Resize(int width, int height)
		{
			//縦横比を保ち、幅を合わせる
			if (height == 0)
			{
				height = (int)((double)img.Height * (double)width / (double)img.Width);
			}
			//縦横比を保ち、高さを合わせる
			else if (width == 0)
			{
				width = (int)((double)img.Width * (double)height / (double)img.Height);
			}
			//それ以外の場合…縦横比を無視し、幅と高さを合わせる

			try
			{
				img = new Bitmap(img, width, height);
				ImgW = img.Width;
				ImgH = img.Height;
			}
			catch { }
		}

		//全画素データをColor形式で取得する
		private Color[] GetPixels()
		{
			Color[] colors = new Color[img.Width * img.Height];
			for (int p = 0, y = 0; y < img.Height; y++)
				for (int x = 0; x < img.Width; x++, p++)
					colors[p] = img.GetPixel(x, y);
			return colors;
		}
		
		//Color形式(0xAARRGGBB)を8bitパレット形式(0bRRRGGGBB)に変換する
		//・0bRRRGGGBBのR,G,Bは色成分を直接表すのではなく、パレットのインデックスを指すことに注意。
		private byte[] RgbToPaletteColor(Color[] colors)
		{
			byte[] colorValues = new byte[colors.Length];
			TRgbColor[] rgbColor = new TRgbColor[colors.Length];	//画像保存のため

			//色味を考えず単純にスケーリングして減色する
			//・256階調を均等に8(4)階調に減らした色値をパレットに設定してある。
			//	そのため、下記計算はパレットのインデックスを直接求めることに相当する。
			for (int rr, gg, bb, i = 0; i < colors.Length; i++)
			{
				rr = colors[i].R * 8 / 256;
				gg = colors[i].G * 8 / 256;
				bb = colors[i].B * 4 / 256;
				colorValues[i] = (byte)((rr << 5) | (gg << 2) | bb);
				
				//画像保存のため色値を保持しておく
				rgbColor[i].rr = S1D15G10Ctrl.DEFAULT_PALETTE[rr];
				rgbColor[i].gg = S1D15G10Ctrl.DEFAULT_PALETTE[8 + gg];
				rgbColor[i].bb = S1D15G10Ctrl.DEFAULT_PALETTE[8 + 8 + bb];
			}

			SaveImage(rgbColor);
			return colorValues;
		}
		
		//RgbToPaletteColor()で使用する
		private struct TRgbColor
		{
			public byte rr, gg, bb;
			public byte paIdxR, paIdxG, paIdxB;
		}
		//
		//RgbToPaletteColor()で使用する
		private struct TColorInfo
		{
			public byte value;
			public int count;
			public bool use;
		}
		//
		//フルカラー画像の色データをLCD用256色に変換する
		private byte[] RgbToPaletteColor(Color[] pixels, out byte[] paletteAry)
		{
			TRgbColor[] rgbColor = new TRgbColor[pixels.Length];
			TColorInfo[] infoR = new TColorInfo[16], infoG = new TColorInfo[16], infoB = new TColorInfo[16];
			for (int i = 0; i < 16; i++)
			{
				infoR[i].value = infoG[i].value = infoB[i].value = (byte)i;
				infoR[i].count = infoG[i].count = infoB[i].count = 0;
				infoR[i].use = infoG[i].use = infoB[i].use = false;
			}

			for (int i = 0; i < pixels.Length; i++)
			{
				//4096色に減色（R,G,Bとも色値0x00～0x0Fの16階調）
				rgbColor[i].rr = (byte)(pixels[i].R * 16 / 256);
				rgbColor[i].gg = (byte)(pixels[i].G * 16 / 256);
				rgbColor[i].bb = (byte)(pixels[i].B * 16 / 256);

				//その色が使われている画素の個数をカウントする
				infoR[rgbColor[i].rr].count++;
				infoG[rgbColor[i].gg].count++;
				infoB[rgbColor[i].bb].count++;
			}

			//R:G:B=16:16:16階調から8:8:4階調に減色し、パレットを取得する
			byte[] paletteColorR = GradationHalf(ref infoR);
			byte[] paletteColorG = GradationHalf(ref infoG);
			byte[] paletteColorB = GradationQuarter(ref infoB);
			
			//16階調の色値を8(4)階調の色値に置き換える
			ReplaceRgbColor(ref rgbColor, infoR, infoG, infoB);

			//8(4)階調の色値に該当するパレットのインデックスをセットする
			SetPaletteIndex(ref rgbColor, paletteColorR, paletteColorG, paletteColorB);

			//パレットのインデックス値からLCD描画用の色値を合成する
			byte[] colorValues = new byte[pixels.Length];
			for (int i = 0; i < rgbColor.Length; i++)
			{
				colorValues[i] =
					(byte)((rgbColor[i].paIdxR << 5) | (rgbColor[i].paIdxG << 2) | rgbColor[i].paIdxB);
			}
			
			//RGB別パレットをまとめる（配列の単純結合）
			paletteAry = new byte[paletteColorR.Length + paletteColorG.Length + paletteColorB.Length];
			int pidx = 0;
			Array.Copy(paletteColorR, 0, paletteAry, pidx, paletteColorR.Length); pidx += paletteColorR.Length;
			Array.Copy(paletteColorG, 0, paletteAry, pidx, paletteColorG.Length); pidx += paletteColorG.Length;
			Array.Copy(paletteColorB, 0, paletteAry, pidx, paletteColorB.Length);

			SaveImage(rgbColor);
			return colorValues;
		}

		//色数を1/2に減らし、パレットを作成する
		private byte[] GradationHalf(ref TColorInfo[] info)
		{
			byte[] paletteColor = new byte[info.Length / 2];

			byte avgValue;
			for (int pidx = 0, i = 0; i < info.Length; i += 2, pidx++)
			{
				avgValue = GetAvgColor(info[i].value, info[i].count, info[i + 1].value, info[i + 1].count);
				paletteColor[pidx] = avgValue;
				info[i].value = info[i + 1].value = avgValue;
			}

			return paletteColor;
		}
		//
		//色の重み付けを考慮して中間色を求める
		private byte GetAvgColor(byte colorA, int countA, byte colorB, int countB)
		{
			int colorWeight = colorA * countA + colorB * countB;
			int totalCount = countA + countB;
			byte avgValue = (byte)((double)colorWeight / (double)totalCount + 0.5);
			return avgValue;
		}
		
		//色数を1/4に減らし、パレットを作成する
		private byte[] GradationQuarter(ref TColorInfo[] info)
		{
			byte[] paletteColor = GradationHalf(ref info);
			TColorInfo[] info2 = new TColorInfo[paletteColor.Length];
			for (int ci = 0, i = 0; i < info2.Length; i++, ci += 2)
			{
				info2[i].value = paletteColor[i];
				info2[i].count = info[ci].count + info[ci + 1].count;
			}

			paletteColor = GradationHalf(ref info2);
			for (int ci = 0, pidx = 0; pidx < paletteColor.Length; pidx++)
				for (int i = 0; i < 4; i++, ci++)
					info[ci].value = paletteColor[pidx];
			
			return paletteColor;
		}

		//16階調の色値を8(4)階調の色値に置き換える
		private void ReplaceRgbColor(ref TRgbColor[] rgbColor,
			TColorInfo[] infoR, TColorInfo[] infoG, TColorInfo[] infoB)
		{
			for (int i = 0; i < rgbColor.Length; i++)
			{
				rgbColor[i].rr = infoR[rgbColor[i].rr].value;
				rgbColor[i].gg = infoG[rgbColor[i].gg].value;
				rgbColor[i].bb = infoB[rgbColor[i].bb].value;
			}
		}

		//8(4)階調の色値に該当するパレット色のインデックスをセットする
		private void SetPaletteIndex(ref TRgbColor[] rgbColor,
			byte[] paletteColorR, byte[] paletteColorG, byte[] paletteColorB)
		{
			for (int i = 0; i < rgbColor.Length; i++)
			{
				rgbColor[i].paIdxR = IndexOfPalette(paletteColorR, rgbColor[i].rr);
				rgbColor[i].paIdxG = IndexOfPalette(paletteColorG, rgbColor[i].gg);
				rgbColor[i].paIdxB = IndexOfPalette(paletteColorB, rgbColor[i].bb);
			}
		}
		//
		//パレットの中から該当する色を探し、インデックスを返す
		private byte IndexOfPalette(byte[] paletteColor, byte color)
		{
			for (byte i = 0; i < paletteColor.Length; i++)
				if (paletteColor[i] == color) return i;
			return 0;
		}

		//全画素データを8bitパレット形式で取得する（LCD転送用データ）
		//戻り値:	変換された色データ
		//(ref)paletteAry:	パレットデータ
		public byte[] GetPixelArray(out byte[] paletteAry)
		{
			Color[] pixColors = GetPixels();
			byte[] lcdPixels = null;
			if (App.isNormalPalette)
			{
				lcdPixels = RgbToPaletteColor(pixColors);
				paletteAry = S1D15G10Ctrl.DEFAULT_PALETTE;
			}
			else
			{
				lcdPixels = RgbToPaletteColor(pixColors, out paletteAry);
			}
			return lcdPixels;
		}
		
	}
}
