﻿/*	ATtiny2313
	aitendo/NOKIA3300-LCD(S1D15G10) 制御テスト
	『昼夜逆転』工作室	http://jsdiy.web.fc2.com

参照設定に System.Drawing を追加すること。

【更新履歴】
2012/05/14	ver 1.00	初版
*/

using System;
using System.IO;	//File
using System.Diagnostics;	//Stopwatch
using ModifiedSerialPort;

namespace lcdS1D15G10com
{
	//エラーコード
	enum ErrorCode
	{
		NormalTermination = 0,
		COM_Error,
		Error = 255
	};

	//オプションで指定される値
	struct App
	{
		public static string portName = null, imgFile = null, savePath = null;
		public static bool isNormalPalette = false;
	}
	
	class Program
	{
		private const string APP_VER = "v1.00";
		private static MSerialPort comCtrl = null;
		private static S1D15G10Ctrl lcdCtrl = null;
		private static string[] COMPortNameList = null;
		private static Stopwatch sw = new Stopwatch();
		
		static void Main(string[] args)
		{
			//UnhandledExceptionイベント・ハンドラを登録する
			AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(Application_UnhandledException);

			//コマンドラインオプションを取得する
			string arg;
			for (int i = 0; i < args.Length; i++)
			{
				arg = args[i].ToLower();
				if (arg.StartsWith("-com"))
					App.portName = "COM" + args[i].Substring(4);
				else if (arg.EndsWith(".jpg"))
					App.imgFile = arg;
				else if (arg == "-n")
					App.isNormalPalette = true;
				else
					App.savePath = arg.EndsWith(@"\") ? arg : arg + @"\";
			}
			
			//COMポート番号と画像ファイルが指定されなかった場合は終了する
			if (string.IsNullOrEmpty(App.portName) || string.IsNullOrEmpty(App.imgFile))
			{
				COMPortNameList = MSerialPort.GetPortNames();
				ShowHelp();
				AppClose(null, ErrorCode.NormalTermination);
			}

			if (!File.Exists(App.imgFile))
				AppClose("ファイルが見つかりません。: " + App.imgFile, ErrorCode.Error);

			//シリアルポートをオープンする
			//デフォルト設定：COM1, 9600bps, Parity.NONE, Data=8bit, StopBits.One
			//ReadLine()で解釈される改行コード：SerialPort.NewLine="\n"
			comCtrl = new MSerialPort();
			comCtrl.ReadTimeout = 2000;		//(ms)
			comCtrl.WriteTimeout = 2000;	//(ms)
			comCtrl.BaudRate = 500000;	//AVR 8MHzでのUSART最高速
			comCtrl.PortName = App.portName;
			try { comCtrl.Open(); }
			catch { AppClose(comCtrl.PortName + " open error.", ErrorCode.COM_Error); }

			//デバイスの接続を確認する
			lcdCtrl = new S1D15G10Ctrl();
			if (lcdCtrl.DeviceContact(comCtrl))
				lcdCtrl.SetSerialPort(comCtrl);
			else
				AppClose(comCtrl.PortName + " NG.", ErrorCode.COM_Error);

			//LCDを初期化する
			lcdCtrl.Init();
			lcdCtrl.Clear(0xFF);	//パッと見の動作確認のため、画面を白／黒で塗りつぶしてみる
			lcdCtrl.Clear(0x00);
			
			sw.Reset();
			sw.Start();
			byte[] paletteAry = lcdCtrl.DrawImage(App.imgFile);
			sw.Stop();

			Console.WriteLine("DrawImage({0} palette) {1:f1}sec", App.isNormalPalette ? "normal" : "optimize", sw.Elapsed.TotalSeconds);
			Console.Write("R:"); for (int i = 0; i < 8; i++) Console.Write(" {0:X02}", paletteAry[i]); Console.WriteLine();
			Console.Write("G:"); for (int i = 8; i < 16; i++) Console.Write(" {0:X02}", paletteAry[i]); Console.WriteLine();
			Console.Write("B:"); for (int i = 16; i < 20; i++) Console.Write(" {0:X02}", paletteAry[i]); Console.WriteLine();
			
			Console.WriteLine();
			Console.WriteLine("[↑][↓]キーで明るさ変更。その他のキーで終了。");

			ConsoleKey key;
			while (true)
			{
				key = Console.ReadKey(true).Key;
				while (Console.KeyAvailable) Console.ReadKey(true);		//キーバッファクリア
				if (key == ConsoleKey.UpArrow) lcdCtrl.SetBrightness(1);
				else if (key == ConsoleKey.DownArrow) lcdCtrl.SetBrightness(-1);
				else break;
			}

			//アプリ正常終了（COMポートのクローズ処理込み）
			AppClose(null, ErrorCode.NormalTermination);
		}
		//<- end of Main()
		
		//未処理の例外をキャッチする
		static void Application_UnhandledException(object sender, UnhandledExceptionEventArgs e)
		{
			Console.WriteLine("予期しないエラーが発生しました。アプリを終了します。");
		}

		//アプリ終了メッセージ
		private static void AppClose(string mes, ErrorCode code)
		{
			//終了メッセージを表示する
			if (mes != null) Console.WriteLine(mes + Environment.NewLine);
			
			//COMポート クローズ
			try { comCtrl.Close(); }
			catch { }

			Console.WriteLine("Press any key.");
			Console.ReadKey();
			
			Environment.Exit((int)code);
		}

		//ヘルプ画面
		private static void ShowHelp()
		{
			Console.WriteLine("NOKIA3300-LCD(S1D15G10) Control Test {0:s}", APP_VER);
			Console.WriteLine("(C)jsdiy/2012~  http://jsdiy.web.fc2.com");
			Console.WriteLine();
			Console.WriteLine(
				"[Usage]" + Environment.NewLine +
				">lcdS1D15G10com.exe imagefile.jpg -com<n> [-n] [<savepath>]" + Environment.NewLine +
				Environment.NewLine +
				"[Options]" + Environment.NewLine +
				"-com<n>    : (必須) COMポート番号  例: -com2" + Environment.NewLine +
				"-n         : (任意) 標準パレットを使用する（パレットを最適化しない）" + Environment.NewLine +
				"<savepath> : (任意) 減色画像の保存パス  例: c:\\mypic"
				);
			Console.WriteLine();
			Console.WriteLine("[Serial port]");
			for (int i = 0; i < COMPortNameList.Length; i++)
				Console.Write(COMPortNameList[i] + " ");
			Console.WriteLine();
			
			Console.WriteLine();
		}

	}	//<- end of class

}	//<- end of namespace
