﻿using System;
using System.IO;	//MemoryStream
using System.Threading;		//Sleep
using ModifiedSerialPort;

namespace lcdS1D15G10com
{
	//パラレルプログラミングのコマンド（文字）
	enum PPCmd
	{							//引数
		LCD_CMD			= '#',	//LCDコマンド[1バイト]
		LCD_DATA		= 'a',	//LCDデータ[1バイト]（数十回の繰り返しを含む）※'a'から連続するn個の文字を使用する
		LCD_DATA_RLEB	= '$',	//LCDデータ[1バイト], 繰り返し回数[1バイト]
		LCD_DATA_RLEW	= '%',	//LCDデータ[1バイト], 繰り返し回数[1ワードHi][Lo]
		WHO				= '@',	//なし
		
		//解釈できないコマンドを受信した場合に送り返す文字（デバイスが）
		RET_UNKNOWN_CMD	= '?'
	};
	
	//S1D15G10のコマンド
	struct LCDCmd
	{
		public const byte
			DISON	= 0xAF,	//Display on
			DISOFF	= 0xAE,	//Display off
			DISNOR	= 0xA6,	//Normal display
			DISINV	= 0xA7,	//Inverse display
			COMSCN	= 0xBB,	//Common scan direction
			DISCTL	= 0xCA,	//Display control
			SLPIN	= 0x95,	//Sleep in
			SLPOUT	= 0x94,	//Sleep out
			PASET	= 0x75,	//Page address set
			CASET	= 0x15,	//Column address set
			DATCTL	= 0xBC,	//Data scan direction, etc.
			RGBSET8	= 0xCE,	//256-color position set
			RAMWR	= 0x5C,	//Writing to memory
			RAMRD	= 0x5D,	//Reading from memory
			PTLIN	= 0xA8,	//Partial display in
			PTLOUT	= 0xA9,	//Partial display out
			RMWIN	= 0xE0,	//Read and modify write
			RMWOUT	= 0xEE,	//End
			ASCSET	= 0xAA,	//Area scroll set
			SCSTART	= 0xAB,	//Scroll start set
			OSCON	= 0xD1,	//Internal oscillation on
			OSCOFF	= 0xD2,	//Internal oscillation off
			PWRCTR	= 0x20,	//Power control
			VOLCTR	= 0x81,	//Electronic volume control
			VOLUP	= 0xD6,	//Increment electronic control by 1
			VOLDOWN	= 0xD7,	//Decrement electronic control by 1
			TMPGRD	= 0x82,	//Temperature gradient set
			EPCTIN	= 0xCD,	//Control EEPROM
			EPCOUT	= 0xCC,	//Cancel EEPROM control
			EPMWR	= 0xFC,	//Write into EEPROM
			EPMRD	= 0xFD,	//Read from EEPROM
			EPSRRD1	= 0x7C,	//Read register 1
			EPSRRD2	= 0x7D,	//Read register 2
			NOP		= 0x25;	//NOP instruction
	};

	class S1D15G10Ctrl
	{
		//COMCTRL_LCD_DATAの'n個'を示す値
		private const int LCD_DATA_CHAR_LENGTH = 30;
		
		//インターフェイスのデバイスに埋め込まれている名前
		private const string DEVICE_NAME = "S1D15G10";
		
		//1回の通信の終了コード
		private const char END_CODE = '\n';
		private static readonly char[] TXD_END_CODE = { END_CODE };
		
		private MSerialPort comCtrl = null;
		public static bool IsComError { get; private set; }	//通信エラーを通知するフラグ

		//NOIKIA3300-LCDの画面サイズ
		private const int
			SCREEN_OFFSET_X = 0,
			SCREEN_OFFSET_Y = 2;
		public const int
			SCREEN_WIDTH = 130,
			SCREEN_HEIGHT = 130;
		
		//NOIKIA3300-LCDの座標系
		//	[0,  0]--[129,  0]		※コントローラの[0,2]がNOIKIA3300-LCD座標系の[0,0]
		//	  :          :
		//	[0,129]--[129,129]
		

		//コンストラクタ
		public S1D15G10Ctrl()
		{
			IsComError = false;
		}
		
		//COMポートオブジェクトをセットする
		public void SetSerialPort(MSerialPort msp)
		{
			comCtrl = msp;
		}

		//指定されたCOMポートにデバイスが接続されているか確認する
		public bool DeviceContact(MSerialPort testComPort)
		{
			if (testComPort == null || !testComPort.IsOpen) return false;
			MSerialPort orgComCtrl = comCtrl;
			comCtrl = testComPort;

			Command(PPCmd.WHO);
			string s = ReadString();
			bool isOK = (!IsComError && s == DEVICE_NAME);

			comCtrl = orgComCtrl;
			return isOK;
		}
		
		//引数の下位4bitの数値0～15にゲタ'A'を履かせて文字に変換する
		private byte itoc(int n) { return (byte)('A' + (n & 0x0F)); }

		//ゲタ'A'を履かせた文字から元の数値0～15に変換する
		private byte ctoi(byte c) { return (byte)(c - 'A'); }

		//コマンドとパラメータを送信する
		//・第2引数は可変長配列。または省略可。
		//・データは、1byteの値を上位／下位4bitに分け、それを加工した文字として送る。
		private void Command(PPCmd cmd, params byte[] values)
		{
			IsComError = false;
			byte[] byteArray = null;
			
			using (MemoryStream memStream = new MemoryStream())
			{
				memStream.WriteByte((byte)cmd);
				
				for (int i = 0; i < values.Length; i++)
				{
					memStream.WriteByte(itoc(values[i] >> 4));
					memStream.WriteByte(itoc(values[i]));
				}
				byteArray = memStream.ToArray();
			}
			GC.Collect();

			try
			{
				comCtrl.ReadExisting();	//受信に備え、PC側の受信バッファをクリアしておく
				comCtrl.Write(byteArray, 0, byteArray.Length);
				comCtrl.Write(TXD_END_CODE, 0, TXD_END_CODE.Length);
			}
			catch { IsComError = true; }
		}
		
		//送信結果を文字列として受信する
		private string ReadString()
		{
			IsComError = false;
			string s = null;
			try { s = comCtrl.ReadLine().Trim(); }	//Trim()は保険
			catch { IsComError = true; }	//タイムアウト発生はキャッチされる
			return s;
		}
		
		//書き込み系コマンドに対するデバイスからの返信の妥当性をチェックする
		//・書き込み系コマンドを送信すると、コマンド自身を表す文字が返ってくる。
		private bool WriteCommandIsDone(PPCmd cmd)
		{
			string sRetData = ReadString();
			if (IsComError) return false;
			string sWrCmd = ((char)cmd).ToString();
			return (sWrCmd == sRetData);
		}

		//LCDコマンドを送信する
		//・第2引数は可変長配列。または省略可。
		private bool SendLcdCommand(byte cmd, params byte[] values)
		{
			byte[] args = { cmd };
			Command(PPCmd.LCD_CMD, args);
			bool isOK = WriteCommandIsDone(PPCmd.LCD_CMD);
			
			for (int i = 0; i < values.Length && isOK; i++)
				isOK = isOK && SendLcdData(values[i]);
			
			return isOK;
		}

		//LCDデータを送信する（1バイトのみ）
		private bool SendLcdData(byte data)
		{
			return SendLcdData(data, 1);
		}
		
		//LCDデータを送信する（RLE形式）
		private bool SendLcdData(byte data, int size)
		{
			bool isOK;
			if (size <= 0 || 0xFFFF < size)
			{
				isOK = false;
			}
			else if (size <= LCD_DATA_CHAR_LENGTH)
			{
				byte[] args = { data };
				Command(PPCmd.LCD_DATA + size - 1, args);
				isOK = WriteCommandIsDone(PPCmd.LCD_DATA_RLEB);	//デバイス側ではLCD_DATA_RLEBとして処理している
			}
			else if (size <= 0xFF)
			{
				byte[] args = { data, (byte)size };
				Command(PPCmd.LCD_DATA_RLEB, args);
				isOK = WriteCommandIsDone(PPCmd.LCD_DATA_RLEB);
			}
			else
			{
				byte[] args = { data, (byte)((size >> 8) & 0xFF), (byte)(size & 0xFF) };
				Command(PPCmd.LCD_DATA_RLEW, args);
				isOK = WriteCommandIsDone(PPCmd.LCD_DATA_RLEW);
			}
			return isOK;
		}

		//パレットの値を設定する
		public void SetColorPalette(byte[] paletteAry)
		{
			if (paletteAry == null || paletteAry.Length != 8 + 8 + 4) return;
			
			SendLcdCommand(LCDCmd.RGBSET8);
			for (int i = 0; i < paletteAry.Length; i++)
				SendLcdData(paletteAry[i]);
		}

		//このアプリでデフォルトとするパレットの値
		public static readonly byte[] DEFAULT_PALETTE =
		{
			/*	このアプリにおけるデフォルト値を適当に決める。
			・4096色(R:G:B=16:16:16階調)中の256色(R:G:B=8:8:4階調)が使用できる。
			→16階調の色値(0x00-0x0F)をR:G:B=8:8:4階調で表現するよう値を割り当てる。
			・画像の内容によって偏った値を振り分けるのも有効。
			→明るい赤の成分を含まない画像なら、赤の8階調を{0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07}とするなど。
			
			下記の定義において、最も暗い色はR,G,B=1,1,2の「黒寸前の灰色」であり、
			「純粋な黒」(R,G,B=0,0,0)が定義されていないパレットの例となっている。
			最も明るい色はR,G,B=15,15,15であり、これは「純粋な白」。
			*/
			0x01, 0x03, 0x05, 0x07, 0x09, 0x0B, 0x0D, 0x0F,		//Red
			0x01, 0x03, 0x05, 0x07, 0x09, 0x0B, 0x0D, 0x0F,		//Green
			0x02, 0x06, 0x0B, 0x0F	//Blue
		};
		
		//NOIKIA3300-LCDの初期化
		public void Init()
		{
			//<Display Setting>
			SendLcdCommand(LCDCmd.DISCTL, 0x00, 0x20);	//SLPOUT前に実行する必要がある
			//引数1:	CL dividing ratio, F1 and F2 drive pattern. default:0x00
			//引数2:	(Numbers of display lines)/4-1 = 132/4-1 = 32 (0x20)
			
			SendLcdCommand(LCDCmd.COMSCN, 0x01);	//コントローラとLCDパネルの配線の対応
			SendLcdCommand(LCDCmd.TMPGRD, 0x00);	//Temperature Gradient（LCDの温度特性）
			SendLcdCommand(LCDCmd.OSCON);	//Internal oscillation on.
			SendLcdCommand(LCDCmd.SLPOUT);	//Sleep out.

			//<Power Supply Setting>
			SendLcdCommand(LCDCmd.VOLCTR, 0x2E, 0x02);	//他に例えば(0x20, 0x03)でも同程度の明るさ
			//引数1:	[6bit] V1 volume value "alpha". default:0x00 (alpha=0)
			//引数2:	[3bit] resistance ratio of the internal resistor. (1+Rb/Ra) default:0x00 (1+2.95)
			//'引数2'で大幅にレンジを変える。'引数1'で小幅に調整する。という使い方。

			SendLcdCommand(LCDCmd.PWRCTR, 0x0F);	//Power control set.

			//<Display Setting 2>
			SendLcdCommand(LCDCmd.DISINV);	//NOIKIA3300-LCDの場合、DISNORでは表示が出ない
			//SendLcdCommand(LCDCmd.PTLOUT);	//デフォルトでPTLOUTなので省略	//Partial out. Exit from the partial display mode.
			//SendLcdCommand(LCDCmd.ASCSET);	//省略
			//SendLcdCommand(LCDCmd.SCSTART);	//省略

			//<Display Setting 3>
			SendLcdCommand(LCDCmd.DATCTL, 0x02, 0x01, 0x01);	//描画データの転送パターンを設定する
			//引数1:	[3bit] 0b00000[P12][P11][P10]
			//	上下反転 P10: Normal/inverse display of the page address. "0": Normal, "1": Inverse
			//	左右反転 P11: Normal/reverse turn of column address. "0": Normal, "1": Reverse
			//	走査方向 P12: Address-scan direction. "0": column direction, "1": page direction
			//	※これらの組み合わせにより、上下/左右反転、90/180/270度回転ができる。
			//引数2:	RGB arrangement.	0x00:0bBBGGGRRR, 0x01:0bRRRGGGBB
			//引数3:	Gray-scale setup.	0x01:8bit-color (0x02:16bit-Type A, 0x04:16bit-TypeB)

			SetColorPalette(DEFAULT_PALETTE);	//LCDCmd.RGBSET8

			//<RAM Setting><RAM Write>
			Clear();

			//<Waiting (approximately 100ms)>
			Thread.Sleep(100);
			SendLcdCommand(LCDCmd.DISON);	//画面表示を出す
		}

		//描画領域を指定する
		//戻り値:	実際にセットされた描画領域のドット数。エラー時は0
		//・(x,y)を起点とし、この点を含み、幅w高さhの長方形領域を指定する。
		//・有効画面領域を超える分はトリミングする。
		private int SetRect(int x, int y, int w, int h)
		{
			if (x < 0 || SCREEN_WIDTH <= x || y < 0 || SCREEN_HEIGHT <= y || w <= 0 || h <= 0) return 0;
			
			if (SCREEN_WIDTH < x + w) w = SCREEN_WIDTH - x;
			if (SCREEN_HEIGHT < y + h) h = SCREEN_HEIGHT - y;
			x += SCREEN_OFFSET_X;
			y += SCREEN_OFFSET_Y;
			
			SendLcdCommand(LCDCmd.PASET, (byte)y, (byte)(y + h - 1));	//(開始座標, 終了座標)
			SendLcdCommand(LCDCmd.CASET, (byte)x, (byte)(x + w - 1));	//(開始座標, 終了座標)
			
			return w * h;
		}

		//指定色で領域を塗りつぶす
		//引数	color: 0bRRRGGGBB
		public void FillRect(int x, int y, int w, int h, byte color)
		{
			int size = SetRect(x, y, w, h);
			if (size == 0) return;
			SendLcdCommand(LCDCmd.RAMWR);
			SendLcdData(color, size);
		}

		//描画を消去する
		//・任意の色で画面全域を塗りつぶす。
		public void Clear() { Clear(0x00); }	//0x00が黒とは限らない。パレットによる。
		public void Clear(byte color)
		{
			FillRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, color);
		}

		//指定色で点を描画する
		//引数	color: 0bRRRGGGBB
		public void SetPixel(int x, int y, byte color)
		{
			FillRect(x, y, 1, 1, color);
		}

		//画像を描画する
		//戻り値:	パレットの配列
		public byte[] DrawImage(string filename)
		{
			if (string.IsNullOrEmpty(filename)) return DEFAULT_PALETTE;
			
			ImageCtrl imgCtrl = new ImageCtrl(filename);
			if (imgCtrl.IsError) return DEFAULT_PALETTE;
			if (imgCtrl.ImgW < imgCtrl.ImgH)
				imgCtrl.Resize(0, SCREEN_HEIGHT);
			else
				imgCtrl.Resize(SCREEN_WIDTH, 0);
			
			byte[] paletteAry;
			byte[] pixels = imgCtrl.GetPixelArray(out paletteAry);
			SetColorPalette(paletteAry);
			
			int posX = (SCREEN_WIDTH - imgCtrl.ImgW) / 2;
			int posY = (SCREEN_HEIGHT - imgCtrl.ImgH) / 2;
			DrawImage(posX, posY, imgCtrl.ImgW, imgCtrl.ImgH, pixels);
			
			return paletteAry;
		}

		//画像を描画する
		//引数	x,y,w,h:	描画領域（画像の縦横サイズと一致していること）
		//		pixels:		配列の要素は0bRRRGGGBB
		private void DrawImage(int x, int y, int w, int h, byte[] pixels)
		{
			if (pixels == null || pixels.Length == 0) return;
			byte[] rleAry = RLECompress(pixels);
			if (rleAry == null) return;

			int size = SetRect(x, y, w, h);
			if (size == 0) return;
			
			SendLcdCommand(LCDCmd.RAMWR);
			for (int i = 0; i < rleAry.Length; i += 3)
				SendLcdData(rleAry[i], (rleAry[i + 1] << 8) | rleAry[i + 2]);
		}
		
		//画像データをRLE圧縮する
		private byte[] RLECompress(byte[] byteAry)
		{
			if (byteAry == null) return null;
			if (byteAry.Length == 0) return null;
			if (byteAry.Length == 1) return new byte[] { byteAry[0], 0, 1 };
			
			byte[] rleAry = null;
			using (MemoryStream memStream = new MemoryStream())
			{
				byte prev = byteAry[0];
				int count = 1;
				for (int i = 1; i < byteAry.Length; i++)
				{
					if (prev == byteAry[i])
					{
						count++;
					}
					else
					{
						memStream.WriteByte(prev);
						memStream.WriteByte((byte)((count >> 8) & 0xFF));
						memStream.WriteByte((byte)(count & 0xFF));
						prev = byteAry[i];
						count = 1;
					}
				}
				memStream.WriteByte(prev);
				memStream.WriteByte((byte)((count >> 8) & 0xFF));
				memStream.WriteByte((byte)(count & 0xFF));
				
				rleAry = memStream.ToArray();
			}
			
			return rleAry;
		}

		//画面の明るさを変更する
		//引数:	変化量を整数で
		public void SetBrightness(int volume)
		{
			SendLcdCommand(0 < volume ? LCDCmd.VOLUP : LCDCmd.VOLDOWN);
		}
		
	}	//<- end of class

}	//<- end of namespace
